#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <time.h>

#include "core.h"
#include "socket.h"
#include "timer.h"
#include "db.h"

#define maxMacAddr 8

#define MAJOR_VERSION (int)0
#define MINOR_VERSION (int)7

struct USER_INFO {
	clock_t last;
	char szNickName[30];
	int majorVersion;
	int minorVersion;
	char szSSID[33];
	char szUID[33];
	int pid;
	int pspCount;
	clock_t ping;
};

struct MAC_ADDRESS {
	unsigned char addr[6];
	int active;
};

struct DATA_HEADER {
	unsigned char dtype;
	unsigned char doption;
	short dsize;
	long sendCount;
	long recvCount;
};

struct client_session_data {
	struct {
		int server_connect: 1;
	} state;
	long sendCount;
	long recvCount;

	int ping_highreso;
	struct MAC_ADDRESS mac[maxMacAddr];
	struct USER_INFO info;
};

int listen_port;
unsigned char *eof="<E>";

int set_count_eof(int fd)
{
	struct client_session_data *csd;
	int i;

	csd = session[fd]->session_data;
	csd->sendCount++;
	WFIFOL(fd,4) = csd->sendCount;
	WFIFOL(fd,8) = csd->recvCount;
	memcpy(WFIFOP(fd,sizeof(struct DATA_HEADER)+WFIFOW(fd,2)),eof,sizeof(eof));
//	printf("send set %d %c %d %d %d %d :",fd,WFIFOB(fd,0),WFIFOB(fd,1),WFIFOW(fd,2),WFIFOL(fd,4),WFIFOL(fd,8));
//	for(i=0;i<WFIFOW(fd,2);i++)
//		printf(" %02x",WFIFOB(fd,sizeof(struct DATA_HEADER)+i));
//	printf("\n");

	return 0;
}

int parse_cl(int fd)
{
	int cmd,len;
	struct client_session_data *csd;
	int i,pos;
	struct DATA_HEADER dh;
	struct USER_INFO *ui;
	unsigned char *data;
	int ping;
	int count;
	char commandType[30];
	char tellTo[30];
	char text[4096];
	const unsigned char bc[6] = {0xff,0xff,0xff,0xff,0xff,0xff};

	if(session[fd]->session_data==NULL){
		session[fd]->session_data = calloc(sizeof(struct client_session_data),1);
		if(session[fd]->session_data == NULL){
			printf("parse_cl : out of memory\n");
			exit(1);
		}
	}
	csd = session[fd]->session_data;

	// ڑ؂Ă̂Ōn
	if(session[fd]->eof){
		close(fd);
		delete_session(fd);
		return 0;
	}

	while(RFIFOREST(fd)>=sizeof(struct DATA_HEADER)){
		len=sizeof(struct DATA_HEADER)+RFIFOW(fd,2)+sizeof(eof);
		if(RFIFOREST(fd)<len)
			return 0;
		if(memcmp(RFIFOP(fd,sizeof(struct DATA_HEADER)+RFIFOW(fd,2)),eof,sizeof(eof))){
			printf("eof not found\n");
			RFIFOSKIP(fd,RFIFOREST(fd));
			return 0;
		}
		csd->recvCount++;
		data = RFIFOP(fd,sizeof(struct DATA_HEADER));

		// debug
//		printf("%d %c %d %d %d %d ",fd,RFIFOB(fd,0),RFIFOB(fd,1),RFIFOW(fd,2),RFIFOL(fd,4),RFIFOL(fd,8));
//		for(i=sizeof(struct DATA_HEADER);i<len;i++){
//			printf("%02x ",RFIFOB(fd,i));
//		}
//		printf("\n");
		memcpy(&dh,RFIFOP(fd,0),sizeof(dh));

		switch(cmd=RFIFOB(fd,0)){
		case 'C':
		case 'U':
			dh.doption = fd;
			dh.dtype = cmd=='C' ? 'c' : 'u' ;
			for(i=0;i<fd_max;i++){
				if(session[i] && session[i]->session_data){
					memcpy(WFIFOP(i,0),&dh,sizeof(dh));
					memcpy(WFIFOP(i,sizeof(dh)),data,dh.dsize);
					set_count_eof(i);
					WFIFOSET(i,sizeof(dh)+dh.dsize+sizeof(eof));
				}
			}
			break;

		case 'D':
			data[dh.dsize] = 0;
			sscanf(data, "/%s %s %4096s", commandType, tellTo, text);
			dh.doption = fd;
			dh.dtype = 'c';
			for(i=0;i<fd_max;i++){
				struct client_session_data *csd2;
				if(session[i] && (csd2=session[i]->session_data) && 
				   strcmp(csd2->info.szNickName, tellTo)==0 ){
					snprintf(WFIFOP(fd,sizeof(dh)),4096, "%s << %s", tellTo, text);
					dh.dsize = strlen(WFIFOP(fd,sizeof(dh)));
					memcpy(WFIFOP(fd,0),&dh,sizeof(dh));
					set_count_eof(fd);
					WFIFOSET(fd,sizeof(dh)+dh.dsize+sizeof(eof));

					snprintf(WFIFOP(i,sizeof(dh)),4096, "%s >> %s", csd->info.szNickName, text);
					dh.dsize = strlen(WFIFOP(i,sizeof(dh)));
					memcpy(WFIFOP(i,0),&dh,sizeof(dh));
					set_count_eof(i);
					WFIFOSET(i,sizeof(dh)+dh.dsize+sizeof(eof));
				}
			}
			break;
		case 'Y':
			data[dh.dsize] = 0;
			sscanf(data, "/%s %4096s", commandType, text);
			dh.doption = fd;
			dh.dtype = 'c';
			for(i=0;i<fd_max;i++){
				struct client_session_data *csd2;
				if(session[i] && (csd2=session[i]->session_data) && 
				   strcmp(csd2->info.szSSID, csd->info.szSSID)==0){
					snprintf(WFIFOP(i,sizeof(dh)),4096, "(%s) %s", csd->info.szNickName, text);
					dh.dsize = strlen(WFIFOP(i,sizeof(dh)));
					memcpy(WFIFOP(i,0),&dh,sizeof(dh));
					set_count_eof(i);
					WFIFOSET(i,sizeof(dh)+dh.dsize+sizeof(eof));
				}
			}
			break;

		case 'T':
			dh.doption = fd;
			dh.dtype = 't';
			for(i=0;i<maxMacAddr;i++){
				if (csd->mac[i].active && memcmp(csd->mac[i].addr, &data[6], 6) == 0)
					break;
			}
			if(i==maxMacAddr){
				for(i=0;i<maxMacAddr;i++){
					if(!csd->mac[i].active){
						memcpy(csd->mac[i].addr,&data[6],6);
						csd->mac[i].active=1;
						break;
					}
				}
			}
			for(i=0;i<fd_max;i++){
				struct client_session_data *csd2;
				if(i!=fd && // dont send self
				   session[i] && (csd2=session[i]->session_data) && 
				   strcmp(csd2->info.szSSID, csd->info.szSSID)==0){
					int s=0;
					if(memcmp(bc,data,6)==0)
						s=1;
					else {
						int j;
						for(j=0;j<maxMacAddr;j++){
							if(memcmp(csd2->mac[j].addr,data,6)==0){
								s=1;
								break;
							}
						}
					}
					if(s){
						memcpy(WFIFOP(i,0),&dh,sizeof(dh));
						memcpy(WFIFOP(i,sizeof(dh)),data,dh.dsize);
						set_count_eof(i);
						WFIFOSET(i,sizeof(dh)+dh.dsize+sizeof(eof));
					}
				}
			}
			break;

		case 'S':
		case 'I':
			ping = csd->info.ping;
			memcpy(&csd->info,data,sizeof(struct USER_INFO));
			csd->info.ping = ping;
			if (csd->info.majorVersion != MAJOR_VERSION || csd->info.minorVersion != MINOR_VERSION) {
				dh.dtype = 'c';
				snprintf(WFIFOP(fd,sizeof(dh)),512, "o[Wv܂: T[o[Ver%d.%d NCAgVer%d.%d",
						 MAJOR_VERSION, MINOR_VERSION, csd->info.majorVersion, csd->info.minorVersion);
				dh.dsize = strlen(WFIFOP(fd,sizeof(dh)));
				memcpy(WFIFOP(fd,0),&dh,sizeof(dh));
				set_count_eof(fd);
				WFIFOSET(fd,sizeof(dh)+dh.dsize+sizeof(eof));
			}
			if(cmd=='I'){
				i= dh.doption;
				dh.dtype = 'i';
				dh.doption = fd;
				if(session[i] && session[i]->session_data){
					memcpy(WFIFOP(i,0),&dh,sizeof(dh));
					memcpy(WFIFOP(i,sizeof(dh)),data,dh.dsize);
					set_count_eof(i);
					WFIFOSET(i,sizeof(dh)+dh.dsize+sizeof(eof));
				}
				break;
			}
			// 'S'
			for(i=0,count=0;i<fd_max;i++){
				if(session[i] && session[i]->session_data){
					count++;
				}
			}
			ui=malloc(sizeof(*ui)*count);
			for(i=0,count=0;i<fd_max;i++){
				struct client_session_data *csd2;
				if(session[i] && (csd2=session[i]->session_data)){
					memcpy(ui+count,&csd2->info,sizeof(*ui));
					count++;
				}
			}
			dh.dsize=count*sizeof(*ui);
			for(i=0;i<fd_max;i++){
				if(session[i] && session[i]->session_data){
					WFIFOB(i,0)='a';
					WFIFOB(i,1)=count;
					WFIFOW(i,2)=dh.dsize;
					memcpy(WFIFOP(i,sizeof(dh)),ui,dh.dsize);
					set_count_eof(i);
					WFIFOSET(i,sizeof(dh)+dh.dsize+sizeof(eof));
				}
			}
			free(ui);
			break;
		case 'P':	// ping
			memcpy(WFIFOP(fd,0),RFIFOP(fd,0),len);
			WFIFOB(fd,0)='p';
			set_count_eof(fd);
			WFIFOSET(fd,len);
			break;

		case 'p':	// pong
// printf("%d %d ->",fd,csd->ping_highreso);
			csd->ping_highreso = (csd->ping_highreso*3 +(gettick_highreso()- RFIFOL(fd,12)) + 2) / 4 ;
// printf(" %d(%d)\n",csd->ping_highreso,gettick_highreso()-RFIFOL(fd,12));
			csd->info.ping = csd->ping_highreso/1000;
		case 'A':	// request 'a'
			for(i=0,count=0;i<fd_max;i++){
				if(session[i] && session[i]->session_data){
					count++;
				}
			}
			ui=malloc(sizeof(*ui)*count);
			for(i=0,count=0;i<fd_max;i++){
				struct client_session_data *csd2;
				if(session[i] && (csd2=session[i]->session_data)){
					memcpy(ui+count,&csd2->info,sizeof(*ui));
					ui[count].ping = (csd->ping_highreso + csd2->ping_highreso)/1000;
					count++;
				}
			}
			dh.dsize=count*sizeof(*ui);

			WFIFOB(fd,0)='a';
			WFIFOB(fd,1)=count;
			WFIFOW(fd,2)=dh.dsize;
			memcpy(WFIFOP(fd,sizeof(dh)),ui,dh.dsize);
			set_count_eof(fd);
			WFIFOSET(fd,sizeof(dh)+dh.dsize+sizeof(eof));

			free(ui);

			break;

		default:
			break;
		}
		RFIFOSKIP(fd,len);
	}

	return 0;
}

int send_ping(void)
{
	int i;

	for(i=0;i<fd_max;i++){
		if(session[i] && session[i]->session_data){
			WFIFOB(i,0)='P';
			WFIFOB(i,1)=0;
			WFIFOW(i,2)=4;
			WFIFOL(i,12)=gettick_highreso();
			set_count_eof(i);
			WFIFOSET(i,12+4+4);
		}
	}
	return 0;
}

int do_init(int argc,char *argv[])
{
	int i;
	char *p;

	if(argc<2){
		printf("%s <listen port>\n",argv[0]);
		exit(1);
	}
	listen_port = atoi(argv[1]);

	if(make_listen_port(listen_port)==0){
		printf("make_listen_port %d\n",listen_port);
		exit(1);
	}

	set_defaultparse(parse_cl);
	add_timer_interval(gettick()+10,send_ping,0,0,2*1000);

	return 0;
}
